//
// Class definitions and Member function definitions
//
//

class identifier_table	// manages the identifier table i.e. variable values
{
 private:
    int tot_var;						// total number of variables until now
    char* var_str[MAX_VAR];             			// variable strings
    int value[MAX_VAR];						// values of variables

    signed int index_of(const char* str)  	// returns array index of variable
	{
	    int found = FALSE;
	    int i;

	    for(i=0; i<tot_var; i++)
		if (!strcmp(var_str[i],str))
		    { found = TRUE; break; }

	    return ( (found) ? i : -1 );
	}

 public:
    identifier_table() {	tot_var = 0; } 	//constructor

    //program interface routines
    int get_tot_vars() 	{ return tot_var; }

    signed int value_of(const char* str)				// return value of a variable
	{
	    signed num = index_of(str);
	    if (num == -1) 
		return -1; 
	    else 
		return value[num];
	}
    
    void set_value(const char* str, int val)		// assign a new value to an old variable
	{
	    signed num = index_of(str);
	    assert(num != -1);
	    value[num] = val;
	}

    void set_value(int index, int val)		// assign a new value to an old variable
	{
	    value[index] = val;
	}

    char* get_var_at(int index)
	{
	    assert(index<tot_var);
	    return (var_str[index]);
	}

    int add_var(const char* str)			// returns array index after
	{
	    assert(tot_var < MAX_VAR);
	    var_str[tot_var] = new char[strlen(str)+1];
	    strcpy(var_str[tot_var], str);
	    value[tot_var] = 0;
	    return (tot_var++);
	}

    void destroy_all()
	{
	    for(int i=0; i<tot_var; i++)
		delete [] var_str[i];
	    tot_var = 0;
	}
};


class input_buffer
/* conveys the input from the user to the program funcs.
		If the input is to be redirected, or if an improvement is aspired in
		the keying in of input from user, modify the routines in this class,
		keeping the program interface routines functionally equivalent       */
{
 private:
    char* inp_ptr;		// ptr to memory containing input string (dynamic alloc)
    int cur_pos;      // current position while reading string thru mem funcs

 public:
    // take input and store it in memory
    void set(const char* str)
	{
	    inp_ptr = new char[strlen(str)+1];
	    strcpy(inp_ptr, str);
	    cur_pos = 0;
	}

    void destroy()	{ delete[] inp_ptr; }

    // program interface routines
    void take_back()  { if (cur_pos > 0) cur_pos--; }

    char next() 
	{ 
	    char to_ret = inp_ptr[cur_pos];
		
	    if (to_ret != 0)
		cur_pos++;

	    return to_ret;
	}
};


class lexical_analyser
/* this class lexically analyses the input buffer and conveys
   the result to the parser */
{
 private:
    input_buffer inp_buf;	//lexical_analyser contains the input buffer
    char string_value[MAX_VAR_LEN+1];	// parser.

    void get_string()   // recognize string in input string
	{
	    char ch;
	    int str_pos = 0;

	    while(1)
		{
		    if ( isalpha( ch=inp_buf.next() ) )
			{
			    if (str_pos==MAX_VAR_LEN)
				error ("Length of variable exceeds 32 characters");
			    else
				string_value[str_pos++] = ch;
			}
		    else
			{
			    string_value[str_pos] = 0;
			    inp_buf.take_back();
			    break;
			}
		}
	}


 public:
    //pass message on to input_buffer
    void set_input(const char* str)	{ return inp_buf.set(str); }
    void destroy() { inp_buf.destroy(); }

    // Program interface routines

    char* read_string() { return string_value; }

    token_types get_token() //main routine
	{
	    char ch;
 
	    while(1)
		{
		    switch(ch = inp_buf.next())
			{
			case '&': 
			    if (inp_buf.next() == '&')
				return AND;
			    else
				error ("Invalid character encountered");

			case '|': 
			    if (inp_buf.next() == '|')
				return OR;
			    else
				error ("Invalid character encountered");
			    
			case '~': return NOT;
			case '(': return LP;
			case ')': return RP;
			
			case 0:
			case 26:
			case 13:    
			case 10:
			    return EOS;

			case 9: //tab
			case 32: //space
			    break;

			default:
			    if (isalpha(ch))
				{
				    inp_buf.take_back();
				    get_string();
				    return VARIABLE;
				}
			    else
				error("Invalid character encountered.");
			}
		}
	}
};

/**** parser class member function defintions ****/

class parser
{    // THE MAIN PARSER //
 private:
    token_types last_token;		//helps cooperation betn member funcs
    int in_paranthesis;	//keeps track of number of para.. we are in
    int to_take;
    char postfix[MAX_EXPR_LEN+1];

    lexical_analyser lex;  // takes help from lexical_analyser
    identifier_table tab;  // and identifier table

    void factor();	// takes care of bracket and tilde
    void term();     // takes care of &&
    void expression(); // takes care of ||

 public:
    //take message onto lexical_analyser
    void set_input(const char* str) 
	{ 
	    in_paranthesis = 0; 
	    to_take = TRUE; 
	    postfix[0] = 0; 
	    lex.set_input(str); 
	}

    void destroy()	{ lex.destroy(); tab.destroy_all(); }
	
    char* get_postfix() { return postfix; }

    //program interface routines
    void convert_postfix() { expression(); }

    int get_num_vars() { return tab.get_tot_vars(); }

    identifier_table* get_table() { return &tab; }
};
	
void parser::expression()
{
    term();

    if (to_take) last_token = lex.get_token();
    to_take = FALSE;

    switch(last_token)
	{
	case OR:
	    {
		do 
		    {
			to_take = TRUE; /* we eat up this OR */
			term();
			strcat(postfix, " ||");

			// check for repetitive ORs
			if (to_take) last_token = lex.get_token();
			to_take = FALSE;
		    }
		while (last_token == OR); // another OR
		
		break;
	    }

	case RP:  		
	    if (!in_paranthesis) error("Unexpected Right Paranthesis.");

	default:
	    {
	    }
	}
}

void parser::term()
{
    factor();

    if (to_take) last_token = lex.get_token();
    to_take = FALSE;

    switch(last_token)
	{
	case AND:
	    {
		do 
		    {
			to_take = TRUE; /* we eat up this AND */
			factor();
			strcat(postfix, " &&");


			// check for repetitive ANDs
			if (to_take) last_token = lex.get_token();
			to_take = FALSE;
		    }
		while (last_token == AND); // another and
		
		break;
	    }

	case RP: 
	    if (!in_paranthesis) error("Unexpected Right Paranthesis.");

	default:
	    {
	    }
	}
}

void parser::factor()
{
    if (to_take) last_token = lex.get_token();
    to_take = TRUE; // we will always gobble up one token

    switch(last_token)
	{
	case NOT:
	    {
		factor();
		strcat (postfix, " ~");
		break;
	    }

	case VARIABLE:
	    {
		char* var = lex.read_string();
		    
		if (tab.value_of(var) == -1)
		    tab.add_var(var);

		strcat (postfix, " ");
		strcat (postfix, var);
		break;
	    }

	case LP:
	    {
		in_paranthesis++;
		expression();
		
		if (to_take) last_token = lex.get_token();
		to_take = TRUE;

		if (last_token == RP)
		    in_paranthesis--; 
		else
		    error("Missing Right Parenthesis");
		break;
	    }

	case EOS:
	    {
		error("Factor expected at end of expr.");
		break;
	    }

	default: error(" Primary Failure.");
	}

};

class postfix 
{
 private:
    int stack[STACK_SIZE];
    int stack_top;
    lexical_analyser lex;

 public:
    postfix() { stack_top = 0; }

    int evaluate(char* str, identifier_table* tab)
	{
	    int to_return;

	    token_types last_token;

	    stack_top = 0;

	    lex.set_input(str);

	    while ((last_token = lex.get_token()) != EOS)
		{
		    switch(last_token)
			{
			case VARIABLE:
			    {
				push(tab->value_of(lex.read_string()));
				break;
			    }

			case AND:
			    {
				int a = pop();
				int b = pop();
				push (a & b);
				break;
			    }
			    
			case OR: 
			    {
				int a = pop();
				int b = pop();
				push (a | b);
				break;
			    }

			case NOT: 
			    {
				int a = pop();
				if (a==0) push(1); else push(0);
				break;
			    }

			default:
			    {
				assert(0);
			    }
			}
		}

	    lex.destroy();

	    to_return = pop();

	    assert (is_empty());
	    
	    return to_return;
	}

    void push (int val) { assert(stack_top < STACK_SIZE); stack[stack_top++] = val; }
    int pop () { assert (stack_top != 0); return stack[--stack_top]; }
    int is_empty() { return ( (stack_top == 0)? 1 : 0 ); }
};
